@@ -13,7 +13,11 @@ from contextlib import ContextDecorator, contextmanager
 from pyrsistent import pmap, m
 
 from django.http.response import HttpResponse, HttpResponseRedirect
-from django.utils.encoding import force_text
+try:
+    from django.utils.encoding import force_text
+except ImportError:
+    from django.utils.encoding import force_str as force_text
+
 from django.utils.safestring import mark_safe
 from django.utils.dateparse import parse_date, parse_datetime, parse_time
 from django.conf import settings
@@ -35,7 +39,7 @@ __all__ = [
     "script", "section", "select", "small", "source", "span", "strike", "strong", "style", "sub", "summary", "sup",
     "svg", "table", "tbody", "td", "template", "textarea", "tfoot", "th", "thead", "time", "title", "tr", "track",
     "tt", "u", "ul", "var", "video", "wbr", "component", "hypergen", "command", "raw", "callback", "call_js", "THIS",
-    "OMIT", "context"]
+    "OMIT", "context", "is_ajax", "write"]
 
 ### Python 2+3 compatibility ###
 
@@ -115,6 +119,12 @@ def wrap2(f):
 def insert(source_str, insert_str, pos):
     return ''.join((source_str[:pos], insert_str, source_str[pos:]))
 
+def is_ajax(request=None):
+    if request is None:
+        request = c.request
+
+    return request.META.get('HTTP_X_REQUESTED_WITH', None) == 'XMLHttpRequest'
+
 ### Rendering ###
 
 def default_wrap_elements(init, self, *children, **attrs):
@@ -154,11 +164,11 @@ def hypergen(func, *args, **kwargs):
         if c.hypergen.event_handler_callbacks:
             command("hypergen.setClientState", 'hypergen.eventHandlerCallbacks', c.hypergen.event_handler_callbacks)
 
-        if not c.request.is_ajax() and c.hypergen.translate and c.user.has_perm("hypergen.kv_hypergen_translations"):
+        if not is_ajax() and c.hypergen.translate and c.user.has_perm("hypergen.kv_hypergen_translations"):
             from hypergen.views import translate
             command("translations", translate.reverse(), [[k, v] for k, v in TRANSLATIONS.items()])
 
-        if not c.request.is_ajax():
+        if not is_ajax():
             pos = html.find("</head")
             if pos != -1:
                 s = "<script type='application/json' id='hypergen-apply-commands-data'>{}</script><script>ready(() => window.applyCommands(JSON.parse(document.getElementById('hypergen-apply-commands-data').textContent, reviver)))</script>".format(
@@ -167,7 +177,8 @@ def hypergen(func, *args, **kwargs):
             print(("Execution time:", (time.time() - a) * 1000, "ms"))
             return html
         else:
-            command("hypergen.morph", c.hypergen.target_id, html)
+            if c.hypergen.target_id != OMIT:
+                command("hypergen.morph", c.hypergen.target_id, html)
             print(("Execution time:", (time.time() - a) * 1000, "ms"))
             return c.hypergen.commands
 
@@ -185,20 +196,20 @@ def hypergen_to_string(func, *args, **kwargs):
 def hypergen_response(html_or_commands_or_http_response, status=None):
     value = html_or_commands_or_http_response
     if isinstance(value, HttpResponseRedirect):
-        if c.request.is_ajax():
+        if is_ajax():
             return HttpResponse(dumps([["hypergen.redirect", value["Location"]]]), status=status,
                 content_type='application/json')
         else:
             return value
     elif isinstance(value, HttpResponse):
         assert status is None
-        assert not c.request.is_ajax()
+        assert not is_ajax()
         return value
     elif type(value) in (list, tuple):
-        assert c.request.is_ajax()
+        assert is_ajax()
         return HttpResponse(dumps(value), status=status, content_type='application/json')
     elif type(value) in (str, str):
-        assert not c.request.is_ajax()
+        assert not is_ajax()
         return HttpResponse(value, status=status)
     else:
         raise Exception("Invalid response value: {}".format(repr(value)))
@@ -254,6 +265,9 @@ def join_html(html):
 def raw(*children):
     c.hypergen.into.extend(children)
 
+def write(*children):
+    c.hypergen.into.extend(t(x) for x in children)
+
 def t(s, quote=True, translatable=False):
     return translate(escape(make_string(s), quote=quote), translatable=translatable)
 
@@ -362,7 +376,7 @@ class THIS(object):
 NON_SCALARS = set((list, dict, tuple))
 DELETED = ""
 
-COERCE = {str: None, str: None, int: "hypergen.coerce.int", float: "hypergen.coerce.float"}
+COERCE = {str: "hypergen.coerce.str", int: "hypergen.coerce.int", float: "hypergen.coerce.float"}
 
 class base_element(ContextDecorator):
     void = False
@@ -432,7 +446,10 @@ class base_element(ContextDecorator):
                 name, a, kw = v.hypergen_callback_signature
                 return "{}({})".format(name, signature(a, kw))
             elif callable(v):
-                return ".".join((v.__module__, v.__name__)).replace("builtins.", "")
+                if hasattr(v, "__name__"):
+                    return ".".join((v.__module__, v.__name__)).replace("builtins.", "")
+                else:
+                    return repr(v)
             elif type(v) is str:
                 return '"{}"'.format(v)
             else:
@@ -813,6 +830,14 @@ class wbr(base_element_void): pass
 # yapf: enable
 
 # Serialization
+ENCODINGS = {
+    datetime.date: lambda o: {"_": ["date", str(o)]},
+    datetime.datetime: lambda o: {"_": ["datetime", str(o)]},
+    tuple: lambda o: {"_": ["tuple", list(o)]},
+    set: lambda o: {"_": ["set", list(o)]},
+    frozenset: lambda o: {"_": ["frozenset", list(o)]},
+    range: lambda o: {"_": ["range", [o.start, o.stop, o.step]]},}
+
 def encoder(o):
     assert not hasattr(o, "reverse"), "Should not happen"
     if issubclass(type(o), base_element):
@@ -821,15 +846,21 @@ def encoder(o):
     elif hasattr(o, "__weakref__"):
         # Lazy strings and urls.
         return make_string(o)
-    if type(o) is datetime.date:
-        return {"_": ["date", str(o)]}
+    fn = ENCODINGS.get(type(o), None)
+    if fn:
+        return fn(o)
     else:
         raise TypeError(repr(o) + " is not JSON serializable")
 
-def dumps(data, default=encoder, indent=None):
-    result = json.dumps(data, default=encoder, separators=(',', ':'), indent=indent)
-
-    return result
+DECODINGS = {
+    "float": float,
+    "date": parse_date,
+    "datetime": parse_datetime,
+    "time": parse_time,
+    "tuple": tuple,
+    "set": set,
+    "frozenset": frozenset,
+    "range": lambda v: range(*v),}
 
 def decoder(o):
     _ = o.get("_", None)
@@ -837,18 +868,18 @@ def decoder(o):
         return o
 
     datatype, value = _
-    if datatype == "float":
-        return float(value)
-    elif datatype == "date":
-        return parse_date(value)
-    elif datatype == "datetime":
-        return parse_datetime(value)
-    elif datatype == "time":
-        return parse_time(value)
+    fn = DECODINGS.get(datatype, None)
+    if fn:
+        return fn(value)
     else:
 
         raise Exception("Unknown datatype, {}".format(datatype))
 
+def dumps(data, default=encoder, indent=None):
+    result = json.dumps(data, default=encoder, separators=(',', ':'), indent=indent)
+
+    return result
+
 def loads(data):
     return json.loads(data, object_hook=decoder)
 
@@ -906,17 +937,29 @@ class Context(threading.local):
     @contextmanager
     def __call__(self, transformer=None, at=None, **items):
         try:
+            # Store previous value.
             ctx = self.ctx
             if at is None:
                 if transformer is not None:
                     self.ctx = transformer(self.ctx)
                 self.ctx = self.ctx.update(m(**items))
             else:
+                if at not in self.ctx:
+                    self.ctx = self.ctx.set(at, pmap(items))
+                else:
+                    new_value_at = self.ctx[at].update(pmap(items))
+                    if not new_value_at:
+                        raise Exception("Not immutable context variable attempted updated. If you want to nest "
+                            "with context() statements you must use a pmap() or another immutable hashmap type.")
+
+                    self.ctx = self.ctx.set(at, new_value_at)
+
                 if transformer is not None:
                     self.ctx = self.ctx.set(at, transformer(self.ctx[at]))
-                self.ctx = self.ctx.set(at, self.ctx[at].update(m(**items)))
+
             yield
         finally:
+            # Reset to previous value.
             self.ctx = ctx
 
 context = Context()
@@ -934,5 +977,4 @@ def context_middleware(get_response):
 
 class ContextMiddleware(MiddlewareMixin):
     def process_request(self, request):
-        # TODO. Change to MIDDLEWARE and not MIDDLEWARE_CLASSES
         context.replace(**_init_context(request))
